"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of functions around Revit Arrow heads.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# --------------------------------------------- Arrow heads  ------------------

#: list of built in parameters attached to dimensions containing arrow head ids
from Autodesk.Revit.DB import BuiltInParameter, ElementType, ElementId, FilteredElementCollector

from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value, get_parameter_value_as_element_id

#:  list of built in parameters attached to
#:
#:    - text
#:    - independent tags
#:    - Annotation symbols
#:
#: containing arrow head ids
ARROWHEAD_PARAS_DIM = [
    BuiltInParameter.DIM_STYLE_CENTERLINE_TICK_MARK,
    BuiltInParameter.DIM_STYLE_INTERIOR_TICK_MARK,
    BuiltInParameter.DIM_STYLE_LEADER_TICK_MARK,
    BuiltInParameter.DIM_LEADER_ARROWHEAD,
    BuiltInParameter.WITNS_LINE_TICK_MARK,
]


#: parameter containing the arrowhead id in text types
ARROWHEAD_PARAS_TEXT = [BuiltInParameter.LEADER_ARROWHEAD]


#: list of built in parameters attached to spot dims containing arrow head ids
#: and symbols used
ARROWHEAD_PARAS_SPOT_DIMS = [
    BuiltInParameter.SPOT_ELEV_LEADER_ARROWHEAD,
    BuiltInParameter.SPOT_ELEV_SYMBOL,
]


#: list of built in parameters attached to stair path types containing arrow head ids
ARROWHEAD_PARAS_STAIRS_PATH = [BuiltInParameter.ARROWHEAD_TYPE]


def get_arrow_head_ids_from_type(doc, type_getter, parameter_list):
    """
    Gets all arrow head symbol ids used in dim types in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_getter: Function getting dimension types in the model.
    :type type_getter: func(doc) returns dim types.
    :param parameter_list: Names of parameters to be checked on dim types.
    :type parameter_list: list of str
    :return: List of element ids representing arrow head symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    used_ids = []
    types = type_getter(doc)
    for t in types:
        for p_int in parameter_list:
            id = get_built_in_parameter_value(t, p_int, get_parameter_value_as_element_id)
            if (
                id not in used_ids
                and id != ElementId.InvalidElementId
                and id != None
            ):
                used_ids.append(id)
            break
    return used_ids


def get_arrow_types_in_model(doc):
    """
    Gets all arrow head types in the model.
    TODO: This uses a plain english name comparison to get arrow head types...may not work in non english versions of Revit!
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element types representing arrow heads.
    :rtype: list of Autodesk.Revit.DB.ElementType
    """

    types = []
    similar_types = []
    col = FilteredElementCollector(doc).OfClass(ElementType)
    for c in col:
        if c.FamilyName == "Arrowhead":
            similar_types = c.GetSimilarTypes()
            # filter out any types not in similar list...not sure what these are...
            if c.Id in similar_types:
                types.append(c)
    return types


def get_arrow_type_ids_in_model(doc):
    """
    Gets all arrow type ids in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing arrow types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    arrow_types = get_arrow_types_in_model(doc)
    ids = []
    for at in arrow_types:
        if at.Id not in ids:
            ids.append(at.Id)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit links, CAD links and image links.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
from duHast.Revit.Common.delete import delete_by_element_ids
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Utilities.Objects import result as res
from duHast.Revit.Links.Utility.link_path import get_link_path

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    Element,
    FilteredElementCollector,
    ModelPathUtils,
    RevitLinkInstance,
    RevitLinkType,
)


def get_all_revit_link_instances(doc):
    """
    Gets all Revit link instances in a model.

    Filters by class.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of revit link instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfClass(RevitLinkInstance)
    return collector


def get_all_revit_link_types(doc):
    """
    Gets all Revit link types in a model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of revit link types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfClass(RevitLinkType)
    return collector


def get_revit_link_type_from_instance(doc, link_instance):
    """
    Gets the Revit link type from a given revit link instance.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param link_instance: The link instance the type of is to be returned
    :type link_instance: Autodesk.Revit.DB.RevitLinkInstance

    :return: The matching revit link type.
    :rtype: Autodesk.Revit.DB.RevitLinkType
    """

    revit_link_types = get_all_revit_link_types(doc)
    for lt in revit_link_types:
        if lt.Id == link_instance.GetTypeId():
            return lt


def get_link_docs(doc, link_names_filter=[]):
    """
    Gets the link documents for the specified link names.

    :param doc: The current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param link_names_filter: A list of link names to filter. Default is an empty list.
    :type link_names_filter: list[str]

    :return: A dictionary mapping link names to their respective link documents.
    :rtype: dict[str, Autodesk.Revit.DB.Document]
    """
    link_docs = {}
    link_instances = get_all_revit_link_instances(doc)
    for link in link_instances:
        link_type = get_revit_link_type_from_instance(doc=doc, link_instance=link)
        link_name = Element.Name.GetValue(link_type)
        if link_name in link_names_filter:
            link_doc = link.GetLinkDocument()
            if link_doc is not None:
                link_docs[link_name] = link.GetLinkDocument()
    return link_docs


def delete_revit_links(doc):
    """
    Deletes all revit links in a file.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return:
        Result class instance.

        - .status True if all revit links got deleted successfully. Otherwise False.
        - .message will contain deletion status. On exception it will also include the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    col = FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_RvtLinks)
    ids = get_ids_from_element_collector(col=col)
    # delete all links at once
    return_value = delete_by_element_ids(
        doc, ids, "Deleting Revit links", "Revit link(s)"
    )
    return return_value


def reload_revit_links(
    doc,
    link_locations,
    host_name_formatted,
    do_something_with_link_name,
    workset_config,
):
    """
    Reloads Revit links from a given file location based on the original link type name (starts with comparison)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param link_locations: A list of directories where CAD files can be located.
    :type link_locations: list str
    :param host_name_formatted: Not used yet
    :type host_name_formatted: TBC
    :param do_something_with_link_name: A function which amends the link name prior search for a match in folders.\
        I.e. can be used to truncate the link name i.e. the revision details of a link
    :type do_something_with_link_name: func(str) -> str
    :param workset_config: To use the previously applied workset config use None, otherwise provide custom config.
    :type workset_config: Autodesk.Revit.DB.WorksetConfiguration

    :return: 
        Result class instance.
        
        - .status True if all revit links got reloaded successfully. Otherwise False.
        - .message will contain status of reload and fully qualified file name. On exception it will also include the exception message.
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        # get all revit link types in model
        for p in FilteredElementCollector(doc).OfClass(RevitLinkType):
            link_type_name = do_something_with_link_name(Element.Name.GetValue(p))
            new_link_path = "unknown"
            try:
                new_link_path = get_link_path(link_type_name, link_locations, ".rvt")
                if new_link_path != None:
                    mp = ModelPathUtils.ConvertUserVisiblePathToModelPath(
                        new_link_path
                    )
                    # attempt to reload with worksets set to last viewed
                    # wc = WorksetConfiguration(WorksetConfigurationOption.OpenLastViewed)
                    # however that can be achieved also ... According to Autodesk:
                    # If you want to load the same set of worksets the link previously had, leave this argument as a null reference ( Nothing in Visual Basic) .
                    wc = workset_config()
                    result = p.LoadFrom(mp, wc)
                    # store result in message
                    return_value.append_message(
                        "{} :: {}".format(link_type_name, result.LoadResult)
                    )
                else:
                    return_value.update_sep(
                        False,
                        "{} :: No link path or multiple path found in provided locations".format(
                            link_type_name
                        ),
                    )
            except Exception as e:
                return_value.update_sep(
                    False, "{} :: Failed with exception: {}".format(link_type_name, e)
                )
    except Exception as e:
        return_value.update_sep(False, "Failed with exception: ".format(e))
    return return_value


def reload_revit_links_from_list(
    doc,
    link_types_tob_reloaded,
    link_locations,
    host_name_formatted,
    do_something_with_link_name,
    workset_config,
):
    """
    Reloads Revit links from a given file location based on the original link type name (starts with comparison)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param link_types_tob_reloaded: List of link type elements to be reloaded.
    :type link_types_tob_reloaded: list Autodesk.Revit.DB.RevitLinkType
    :param link_locations: A list of directories where CAD files can be located.
    :type link_locations: list str
    :param host_name_formatted: Not used yet
    :type host_name_formatted: TBC
    :param do_something_with_link_name: A function which amends the link name prior search for a match in folders.\
        I.e. can be used to truncate the link name i.e. the revision details of a link
    :type do_something_with_link_name: func(str) -> str
    :param workset_config: To use the previously applied workset config use None, otherwise provide custom config.
    :type workset_config: Autodesk.Revit.DB.WorksetConfiguration

    :return: 
        Result class instance.
        
        - .status True if all revit links got reloaded successfully. Otherwise False.
        - .message will contain status of reload and fully qualified file name. On exception it will also include the exception message.
    
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        # loop over links supplied
        for p in link_types_tob_reloaded:
            link_type_name = do_something_with_link_name(Element.Name.GetValue(p))
            new_link_path = "unknown"
            try:
                new_link_path = get_link_path(link_type_name, link_locations, ".rvt")
                if new_link_path != None:
                    mp = ModelPathUtils.ConvertUserVisiblePathToModelPath(
                        new_link_path
                    )
                    # attempt to reload with worksets set to last viewed
                    # wc = WorksetConfiguration(WorksetConfigurationOption.OpenLastViewed)
                    # however that can be achieved also ... According to Autodesk:
                    # If you want to load the same set of worksets the link previously had, leave this argument as a null reference ( Nothing in Visual Basic) .
                    wc = workset_config()
                    result = p.LoadFrom(mp, wc)
                    # store result in message
                    return_value.append_message(
                        "{} :: {}".format(link_type_name, result.LoadResult)
                    )
                else:
                    return_value.update_sep(
                        False,
                        "{} :: No link path or multiple path found in provided locations".format(
                            link_type_name
                        ),
                    )
            except Exception as e:
                return_value.update_sep(
                    False, "{} :: Failed with exception: ".format(link_type_name, e)
                )
    except Exception as e:
        return_value.update_sep(
            False, "Reload Revit links Failed with exception: ".format(e)
        )
    return return_value


def default_link_name(name):
    """
    Default 'do something with link name' method. Returns the link name unchanged.

    Could be replaced with something which i.e. truncates the revision...

    :param name: The link name.
    :type name: str

    :return: the link name unchanged.
    :rtype: str
    """

    return name


def default_workset_config_for_reload():
    """
    Default method returning an 'open previous worksets' configuration. (None)

    :return: None: which is use the previous workset configuration.
    :rtype: None
    """

    return None



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit purging not used fill pattern styles using purge by delete helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.LinePattern.fill_patterns import get_all_fill_pattern, get_used_pattern_ids_in_filled_region_types
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Purge.purge_unused_by_delete import purge_unused_elements

from duHast.Utilities.Objects.result import Result


def get_fill_pattern_ids(doc):
    """
    Returns all fill pattern ids in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of all fill pattern ids in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    fill_pattern_col = get_all_fill_pattern(doc)
    ids = get_ids_from_element_collector(fill_pattern_col)

    # reduce list of ids by removing ids used in filled region types
    # get all fill pattern ids used in filled region types
    used_pattern_ids = get_used_pattern_ids_in_filled_region_types(doc)
    # only return ids not used in filled region types
    ids = [id for id in ids if id not in used_pattern_ids]

    return ids

def purge_fill_pattern_by_delete(doc, progress_callback=None, debug=False):
    """
    Purge fill pattern by delete.

    Note: This is a very slow process and can take a few hours to complete depending on the size of the model and the number of fill patterns.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param progress_callback: Callback to report progress.
    :type progress_callback: callable
    :param debug: Debug mode.
    :type debug: bool
    :return: Result class instance.

        - .status True if unused fill pattern where deleted or nothing needed to be deleted. Otherwise False.
        - .message will contain deletion status.

    """

    return_value = Result()

    try:

        # purge unused fill patterns
        purge_result = purge_unused_elements(
            doc=doc,
            element_id_getter=get_fill_pattern_ids,
            deleted_elements_modifier=None,
            modified_elements_modifier=None,
            progress_callback=progress_callback,
            debug=debug,
        )

        return_value.update(purge_result)
    except Exception as e:
        return_value.update_sep(
            False, "Error purging fill pattern by delete: {}".format(str(e))
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of helper functions relating to Revit worksets.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

import System

# import common library modules
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import transaction as rTran
from duHast.Utilities import utility as util
from duHast.Utilities import files_io as filesIO
from duHast.Utilities import files_tab as filesTab

# import Autodesk
import Autodesk.Revit.DB as rdb
from Autodesk.Revit.DB import (
    BuiltInParameter,
    Element,
    ElementId,
    ElementWorksetFilter,
    FilteredElementCollector,
    FilteredWorksetCollector,
    Transaction,
    Workset,
    WorksetKind,
    WorksetDefaultVisibilitySettings,
    WorksetTable,
)

# --------------------------------------------- utility functions ------------------


def get_workset_id_by_name(doc, workset_name):
    """
    Returns the element id of a workset identified by its name, otherwise invalid Id (-1) if no such workset exists

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param workset_name: The name of the workset of which to retrieve the Element Id
    :type workset_name: str

    :return: The workset element id, otherwise invalid Id (-1) if no such workset exists
    :rtype: Autodesk.Revit.DB.ElementId
    """

    id = ElementId.InvalidElementId
    for p in FilteredWorksetCollector(doc).OfKind(WorksetKind.UserWorkset):
        if p.Name == workset_name:
            id = p.Id
            break
    return id


def get_workset_name_by_id(doc, id_integer):
    """
    Returns the name of the workset identified by its Element Id, otherwise 'unknown' if no such workset exists

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param id_integer: The element id as integer value.
    :type id_integer: int

    :return: The name of the workset identified by its Id, otherwise 'unknown'
    :rtype: str
    """

    name = "unknown"
    for p in FilteredWorksetCollector(doc).OfKind(WorksetKind.UserWorkset):
        if p.Id.IntegerValue == id_integer:
            name = p.Name
            break
    return name


def get_workset_ids(doc):
    """
    Gets all ids of all user defined worksets in a model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of all user defined workset element ids
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    id = []
    for p in FilteredWorksetCollector(doc).OfKind(WorksetKind.UserWorkset):
        id.append(p.Id)
    return id


def get_worksets(doc):
    """
    Returns all user defined worksets in the model as list.

    Will return a list of zero length if no worksets in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of worksets
    :rtype: list of Autodesk.Revit.DB.Workset
    """

    worksets = FilteredWorksetCollector(doc).OfKind(WorksetKind.UserWorkset).ToList()
    return worksets


def get_worksets_from_collector(doc):
    """
    Returns all user defined worksets in the model as collector.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered elements collector of user defined worksets
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredWorksetCollector(doc).OfKind(WorksetKind.UserWorkset)
    return collector


def open_worksets_with_elements_hack(doc):
    """
    This is based on a hack from the AutoDesk forum and an article from the building coder:

    - https://forums.autodesk.com/t5/revit-api-forum/open-closed-worksets-in-open-document/td-p/6238121
    - https://thebuildingcoder.typepad.com/blog/2018/03/switch-view-or-document-by-showing-elements.html

    this method will open worksets in a model containing elements only

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    """

    # get worksets in model
    workset_ids = get_workset_ids(doc)
    # loop over workset and open if anything is on them
    for w_id in workset_ids:
        workset = ElementWorksetFilter(w_id)
        elem_ids = FilteredElementCollector(doc).WherePasses(workset).ToElementIds()
        if len(elem_ids) > 0:
            # this will force Revit to open the workset containing this element
            rdb.uidoc.ShowElements(elem_ids.First())


def modify_element_workset(doc, default_workset_name, collector, element_type_name):
    """
    Attempts to change the worksets of elements provided through an element collector.

    Will return false if target workset does not exist in file.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param default_workset_name: The name of the workset the elements are to be moved to.
    :type default_workset_name: str
    :param collector: The element collector containing the elements.
    :type collector: Autodesk.Revit.DB.FilteredElementCollector
    :param element_type_name: A description used in the status message returned.
    :type element_type_name: str

    :return:
        Result class instance.

        - .status = True if successfully moved all elements to new workset. Otherwise False.
        - .message will contain stats in format [success :: failure]

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    return_value.message = "Changing: {} workset to: {} ".format(
        element_type_name, default_workset_name
    )
    # get the ID of the default grids workset
    default_id = get_workset_id_by_name(doc, default_workset_name)
    counter_success = 0
    counter_failure = 0
    # check if invalid id came back..workset no longer exists..
    if default_id != ElementId.InvalidElementId:
        # get all elements in collector and check their workset
        for p in collector:
            if p.WorksetId != default_id:
                # get the element name
                element_name = "Unknown Element Name"
                try:
                    element_name = Element.Name.GetValue(p)
                except Exception:
                    pass
                # move element to new workset
                transaction = Transaction(
                    doc, "Changing workset: ".format(element_name)
                )
                tranny_status = rTran.in_transaction(
                    transaction, get_action_change_element_workset(p, default_id)
                )
                if tranny_status.status == True:
                    counter_success += 1
                else:
                    counter_failure += 1
                return_value.status = return_value.status & tranny_status.status
            else:
                counter_success += 1
                return_value.status = return_value.status & True
    else:
        return_value.update_sep(
            False,
            "Default workset: {} does no longer exists in file!".format(
                default_workset_name
            ),
        )
    return_value.append_message(
        "Moved: {} to workset: {} [ {} :: {}]".format(
            element_type_name, default_workset_name, counter_success, counter_failure
        )
    )
    return return_value


def get_action_change_element_workset(el, default_id):
    """
    Contains the required action to change a single elements workset

    :param el: The element
    :type el: Autodesk.Revit.DB.Element
    :param default_id: The workset element Id
    :type default_id: Autodesk.Revit.DB.ElementId
    """

    def action():
        action_return_value = res.Result()
        try:
            ws_param = el.get_Parameter(BuiltInParameter.ELEM_PARTITION_PARAM)
            ws_param.Set(default_id.IntegerValue)
            action_return_value.message = "Changed element workset."
        except Exception as e:
            action_return_value.update_sep(False, "Failed with exception: {}".format(e))
        return action_return_value

    return action


def is_element_on_workset_by_id(doc, el, workset_id):
    """
    Checks whether an element is on a given workset

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param el: The element.
    :type el: Autodesk.Revit.DB.Element
    :param workset_id: The workset element Id
    :type workset_id: Autodesk.Revit.DB.ElementId

    :return: True if element is on given workset, otherwise False
    :rtype: bool
    """

    flag = True
    try:
        ws_param = el.get_Parameter(BuiltInParameter.ELEM_PARTITION_PARAM)
        current_workset_name = rParaGet.get_parameter_value(ws_param)
        compare_to_workset_name = get_workset_name_by_id(doc, workset_id.IntegerValue)
        if compare_to_workset_name != current_workset_name:
            flag = False
    except Exception as e:
        print("is_element_on_workset_by_id: {}".format(e))
        flag = False
    return flag


def is_element_on_workset_by_name(el, workset_name):
    """
    Checks whether an element is on a workset identified by name

    :param el: The element
    :type el: Autodesk.Revit.DB.Element
    :param workset_name: The name of the workset
    :type workset_name: str

    :return: True if element is on given workset, otherwise False
    :rtype: bool
    """

    flag = True
    try:
        ws_param = el.get_Parameter(BuiltInParameter.ELEM_PARTITION_PARAM)
        current_workset_name = rParaGet.get_parameter_value(ws_param)
        if workset_name != current_workset_name:
            flag = False
    except Exception as e:
        print("IsElementOnWorksetByName: {}".format(e))
        flag = False
    return flag


def get_element_workset_name(el):
    """
    Returns the name of the workset an element is on, or 'invalid workset'.

    :param el: The element.
    :type el: Autodesk.Revit.DB.Element
    :return: The name of the workset. If an exception occurred it wil return 'invalid workset'.
    :rtype: str
    """

    work_setname = "invalid workset"
    try:
        ws_param = el.get_Parameter(BuiltInParameter.ELEM_PARTITION_PARAM)
        work_setname = rParaGet.get_parameter_value(ws_param)
    except Exception as e:
        print("GetElementWorksetName: " + str(e))
    return work_setname


def update_workset_default_visibility_from_report(doc, report_path, revit_file_path):
    """
    Updates the default visibility of worksets based on a workset report file.

    The data for the report files is generated by GetWorksetReportData() function in this module

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param report_path: The fully qualified file path to tab separated report text file containing workset data.
    :type report_path: str
    :param revit_file_path: The fully qualified file path of the Revit file. Will be used to identify the file in the report data.
    :type revit_file_path: str

    :return:
        Result class instance.

        - .status = True if:

            - successfully updated all workset default visibility where different to report
            - or none needed updating.

        - Otherwise False:

            - An exception occurred.
            - A workset has no matching data in the report.

        - Common:

            - .message will contain each workset and whether it needed updating or not

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # read report
    workset_data = filesTab.read_tab_separated_file(report_path)
    file_name = filesIO.get_file_name_without_ext(revit_file_path)
    workset_data_for_file = {}
    for row in workset_data:
        if (
            filesIO.get_file_name_without_ext(row[0]).startswith(file_name)
            and len(row) > 3
        ):
            workset_data_for_file[row[1]] = util.parse_string_to_bool(row[3])
    if len(workset_data_for_file) > 0:
        # updates worksets
        worksets = get_worksets(doc)
        for workset in worksets:
            if str(workset.Id) in workset_data_for_file:
                if workset.IsVisibleByDefault != workset_data_for_file[str(workset.Id)]:

                    def action():
                        action_return_value = res.Result()
                        default_visibility = WorksetDefaultVisibilitySettings.GetWorksetDefaultVisibilitySettings(
                            doc
                        )
                        try:
                            default_visibility.SetWorksetVisibility(
                                workset.Id, workset_data_for_file[str(workset.Id)]
                            )
                            action_return_value.update_sep(
                                True,
                                "{}: default visibility settings changed to: \t[{}]".format(
                                    workset.Name, workset_data_for_file[str(workset.Id)]
                                ),
                            )
                        except Exception as e:
                            action_return_value.update_sep(
                                False, "Failed with exception: {}".format(e)
                            )
                        return action_return_value

                    # move element to new workset
                    transaction = Transaction(
                        doc,
                        "{}: Changing default workset visibility".format(workset.Name),
                    )
                    tranny_status = rTran.in_transaction(transaction, action)
                    return_value.update(tranny_status)
                else:
                    return_value.update_sep(
                        True,
                        "{}: default visibility settings unchanged.".format(
                            util.encode_ascii(workset.Name)
                        ),
                    )
            else:
                return_value.update_sep(
                    False,
                    "{}: has no corresponding setting in settings file.".format(
                        util.encode_ascii(workset.Name)
                    ),
                )
    else:
        return_value.update_sep(
            True, "No settings found for file: {}".format(file_name)
        )
    return return_value


def create_workset(doc, workset_name):
    """
    Creates a workset with a given name in a document.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param workset_name: The name of the workset to be created
    :type workset_name: str

    :return:
        Result class instance.

        - .status = True if:

            - successfully created the workset

        - Otherwise False:

            - An exception occurred.
            - A workset with that name already exists.
            - Document is not workshared

        - Common:

            - .message will contain each workset

        .result will contain the workset element in a list if successfully created, otherwise an empty list.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    if (doc.IsWorkshared):
        if(WorksetTable.IsWorksetNameUnique(doc, workset_name)):
            def action():
                action_return_value = res.Result()
                try:
                    new_workset = Workset.Create(doc, workset_name)
                    action_return_value.update_sep(True, "Workset {} created.".format(workset_name))
                    action_return_value.result.append(new_workset)
                except Exception as e:
                    action_return_value.update_sep(False, "Failed to create workset {} with exception: {}".format(workset_name,e))
                return action_return_value
            
            transaction = Transaction(doc, "Adding workset {}".format(workset_name))
            tranny_status = rTran.in_transaction(transaction, action)
            return_value.update(tranny_status)
        else:
            return_value.update_sep(False, "Workset name already exists: {}".format(workset_name))
    else:
        return_value.update_sep(False, "Cant create a workset in a non workshared file.")
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging annotation types. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    BuiltInParameter,
    ElementId,
    FilteredElementCollector,
    ViewSchedule,
)

from duHast.Revit.Annotation.generic_annotation import (
    get_all_generic_annotation_type_ids_by_category,
)
from duHast.Revit.Family.purge_unused_family_types import (
    get_unused_in_place_ids_for_purge,
)
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
from duHast.Revit.Common.common import get_unused_type_ids_in_model
from duHast.Revit.Annotation.annotation import get_anno_symbol_arrow_head_ids
from duHast.Revit.Annotation.dimensions import (
    get_all_dimension_elements,
    get_dim_type_arrow_head_ids,
    get_dim_types,
    get_symbol_ids_from_dim_types,
)
from duHast.Revit.Annotation.multi_ref_annotation import (
    get_all_multi_ref_annotation_elements,
    get_all_similar_multi_reference_anno_types,
    get_all_multi_ref_annotation_types,
)
from duHast.Revit.Annotation.text import (
    get_all_text_annotation_elements,
    get_text_type_arrow_head_ids,
    get_all_text_types,
)
from duHast.Revit.Annotation.arrow_heads import get_arrow_type_ids_in_model
from duHast.Revit.Annotation.independent_tags import (
    get_independent_tag_type_arrow_head_ids,
)
from duHast.Revit.Annotation.spot_dimensions import (
    get_spot_type_arrow_head_ids,
    get_all_spot_elevation_symbol_ids_in_model,
    get_all_spot_dim_types,
    get_symbol_ids_from_spot_types,
)
from duHast.Revit.Annotation.stair_path import get_stairs_path_arrow_head_ids
from duHast.Revit.Common.purge_utils import get_used_unused_type_ids


# ------------------ used annotation types  ------------------


def get_used_text_type_ids_in_model(doc):
    """
    Gets all ids of text types used by elements in the model, includes types used in schedules (appearance)!
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used text types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    text_type_ids_used = []
    col = get_all_text_annotation_elements(doc)
    for t in col:
        if t.GetTypeId() not in text_type_ids_used:
            text_type_ids_used.append(t.GetTypeId())
    # get all schedules and check their appearance text properties!
    col = FilteredElementCollector(doc).OfClass(ViewSchedule)
    for c in col:
        if c.BodyTextTypeId not in text_type_ids_used:
            text_type_ids_used.append(c.BodyTextTypeId)
        if c.HeaderTextTypeId not in text_type_ids_used:
            text_type_ids_used.append(c.HeaderTextTypeId)
        if c.TitleTextTypeId not in text_type_ids_used:
            text_type_ids_used.append(c.TitleTextTypeId)
    return text_type_ids_used


def get_used_dim_type_ids_in_model(doc):
    """
    Gets all used dimension type Ids in the model.
    Used: at least one instance using this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing Dimension Types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    dim_type_ids_used = []
    col = get_all_dimension_elements(doc)
    for v in col:
        if v.GetTypeId() not in dim_type_ids_used:
            dim_type_ids_used.append(v.GetTypeId())
    return dim_type_ids_used


def get_used_dim_styles_from_multi_ref(doc, multi_reference_anno_types):
    """
    Gets all dimension styles used in multi ref annotation types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param multi_reference_anno_types: list in format [[multi refType, [element ids of similar multi ref types, ...]]]
    :type multi_reference_anno_types: List [[Autodesk.Revit.DB.ElementType, [Autodesk.Revit.DB.ElementId, Autodesk.Revit.DB.ElementId,...],]]
    :return: List of element ids representing dimension style
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    dim_type_ids_used = []
    for m_type in multi_reference_anno_types:
        for t in m_type[1]:
            multi_ref_type = doc.GetElement(t)
            if multi_ref_type.DimensionStyleId not in dim_type_ids_used:
                dim_type_ids_used.append(multi_ref_type.DimensionStyleId)
    return dim_type_ids_used


def get_used_multi_ref_dim_type_ids_in_model(doc):
    """
    Gets all ids of multi reference types used by elements in the model.
    Used: at least one instance using this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing multi reference Annotation Types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    dim_type_ids_used = []
    col = get_all_multi_ref_annotation_elements(doc)
    for v in col:
        if v.GetTypeId() not in dim_type_ids_used:
            dim_type_ids_used.append(v.GetTypeId())
    return dim_type_ids_used


def get_all_used_arrow_head_type_ids_in_model(doc):
    """
    Returns all used arrow types in the model.
    Used in types of dimension, text, independent tags, spot dims, annotation symbols (incl room and area tags), stairs path
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing arrow head symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    used_ids = []
    used_ids = used_ids + get_dim_type_arrow_head_ids(doc)
    used_ids = used_ids + get_text_type_arrow_head_ids(doc)
    used_ids = used_ids + get_independent_tag_type_arrow_head_ids(doc)
    used_ids = used_ids + get_spot_type_arrow_head_ids(doc)
    used_ids = used_ids + get_anno_symbol_arrow_head_ids(doc)
    used_ids = used_ids + get_stairs_path_arrow_head_ids(doc)
    filtered_ids = []
    for u in used_ids:
        if u not in filtered_ids:
            filtered_ids.append(u)
    return filtered_ids


# ------------------ unused annotation types  ------------------


def get_all_unused_text_type_ids_in_model(doc):
    """
    Gets ID of all unused text types in the model.
    Unused: Not one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused text types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    filtered_unused_text_type_ids = get_unused_type_ids_in_model(
        doc, get_all_text_types, get_used_text_type_ids_in_model
    )
    return filtered_unused_text_type_ids


def get_all_unused_dim_type_ids_in_model(doc):
    """
    Gets ID of all unused dim types in the model.
    Includes checking multi ref dims for used dim types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing dimension types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get unused dimension type ids
    filtered_unused_dim_type_ids = get_unused_type_ids_in_model(
        doc, get_dim_types, get_used_dim_type_ids_in_model
    )
    # get all multi ref dimension types in model
    multi_reference_anno_types = get_all_similar_multi_reference_anno_types(doc)
    # get all dim styles used in multi refs
    used_dim_styles_in_multi_refs = get_used_dim_styles_from_multi_ref(
        doc, multi_reference_anno_types
    )
    # cross reference filtered list vs multi ref list and only keep items which are just in the filtered list
    unused_dim_type_ids = []
    for f in filtered_unused_dim_type_ids:
        if f not in used_dim_styles_in_multi_refs:
            unused_dim_type_ids.append(f)
    return unused_dim_type_ids


def get_all_unused_multi_ref_dim_type_ids_in_model(doc):
    """
    Gets IDs of all unused multi ref dimension types in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing multi ref dimension types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    return get_unused_type_ids_in_model(
        doc,
        get_all_multi_ref_annotation_types,
        get_used_multi_ref_dim_type_ids_in_model,
    )


def get_all_unused_arrow_type_ids_in_model(doc):
    """
    Gets all unused arrow type ids in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing arrow types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    unused_ids = []
    used_ids = get_all_used_arrow_head_type_ids_in_model(doc)
    available_ids = get_arrow_type_ids_in_model(doc)
    for a_id in available_ids:
        if a_id not in used_ids:
            unused_ids.append(a_id)
    return unused_ids


def get_unused_symbol_ids_from_spot_types(doc):
    """
    Gets all family symbol ids not used as symbol in any spot elevation or spot coordinate type definition.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing family symbols not used in spot elevation or spot coordinate type definition.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = []
    ids_available = get_all_spot_elevation_symbol_ids_in_model(doc)
    dim_ts = get_all_spot_dim_types(doc)
    for t in dim_ts:
        id = get_built_in_parameter_value(t, BuiltInParameter.SPOT_ELEV_SYMBOL)
        if id not in ids_used and id != ElementId.InvalidElementId and id != None:
            ids_used.append(id)

    # get unused ids
    for id in ids_available:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_symbol_ids_from_spot_types_to_purge(doc):
    """
    Gets all unused family and family symbol ids of category BuiltInCategory.OST_SpotElevSymbols.
    This method can be used to safely delete unused families.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing family and family symbols not used in spot elevation or spot coordinate type definition.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_in_place_ids_for_purge(doc, get_unused_symbol_ids_from_spot_types)
    return ids


def get_used_generic_annotation_type_ids(doc):
    """
    Returns all used generic annotation symbol ids ( used in model as well as dimension types)
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: _description_
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    # get ids from symbols used in dim types
    ids_dim_types = get_symbol_ids_from_dim_types(doc)
    # get ids from symbols used in spots
    ids_spots = get_symbol_ids_from_spot_types(doc)
    # get detail types used in model
    ids_used_in_model = get_used_unused_type_ids(
        doc, get_all_generic_annotation_type_ids_by_category, 1
    )
    # build overall list
    for id in ids_used_in_model:
        ids.append(id)
    for id in ids_dim_types:
        if id not in ids:
            ids.append(id)
    for id in ids_spots:
        if id not in ids:
            ids.append(id)
    return ids


def get_unused_generic_annotation_type_ids(doc):
    """
    Returns all unused annotation symbol ids ( unused in model as well as dimension types)
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: _description_
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = get_used_generic_annotation_type_ids(doc)
    ids_all = get_all_generic_annotation_type_ids_by_category(doc)
    for id in ids_all:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_generic_annotation_ids_for_purge(doc):
    """
    returns symbol(type) ids and family ids (when no type is in use) of in generic anno families which can be purged
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: _description_
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_in_place_ids_for_purge(doc, get_unused_generic_annotation_type_ids)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Customizable element filter class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This class takes as constructor arguments

- a number of filter actions
- a boolean indicating whether filters are logical and or OR filters

Filter actions checks whether a property matches or does not match provided values. Refer to module: custom_element_filter_actions. The actual property test is undertaken by another function. Samples of those can be found in
module custom_element_filter_tests.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import Autodesk
# import Autodesk.Revit.DB as rdb

from duHast.Utilities.Objects import base


class RevitCustomElementFilter(base.Base):
    def __init__(self, element_filters=[], is_logical_and_filter=True, **kwargs):
        """
        Constructor: This takes a list of element filters and a flag whether this class instance is a logical AND filter (default)

        :param element_filters: List of element filter functions which will need to accept document and elementId as their arguments, defaults to []
        :type element_filters: list of functions, optional
        :param is_logical_and_filter: Flag indicating whether list of filters are logical AND filters or logical OR, defaults to True (logical AND)
        :type is_logical_and_filter: bool, optional
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(RevitCustomElementFilter, self).__init__(**kwargs)

        self.element_filters = element_filters
        self.is_logical_and_filter = is_logical_and_filter

    def check_element(self, doc, element_id):
        """
        Filter checking whether element meets criteria.

        This function will loop over all the filters past in through the class constructor and test the element for each filter.
        Depending on whether these filters are logical and filters it will return True if all of them evaluate to True or, if logical or filter
        it will return True if one of them evaluates to True, otherwise False will be returned.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :param element_id: The id of the element to be checked against the filter.
        :type element_id: Autodesk.Revit.DB.ElementId
        :return: True if it matches the filter(s), otherwise False
        :rtype: bool
        """

        if self.is_logical_and_filter:
            filter_over_all = True
        else:
            filter_over_all = False

        for filter in self.element_filters:
            filter_result = filter(doc, element_id)
            if self.is_logical_and_filter == False and filter_result == True:
                filter_over_all = True
                # can leave the loop at this point since only one of the tests need to result true in OR filter mode
                break
            elif self.is_logical_and_filter == True and filter_result == False:
                filter_over_all = False
                # can leave the loop at this point since only one of the tests need to result false in AND filter mode
                break
            else:
                filter_over_all = filter_over_all and filter_result
        return filter_over_all



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit purging not used line styles by using the purge by delete helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.LinePattern.line_styles import get_all_line_style_ids
from duHast.Revit.Purge.purge_unused_by_delete import purge_unused_elements
from duHast.Revit.LinePattern.Objects.LineStylePurgeModifier import (
    LineStylePurgeModifier,
)
from duHast.Utilities.Objects.result import Result


def purge_line_styles_by_delete(doc, progress_callback=None, debug=False):
    """
    Purge line styles by delete.

    Note: This is a quick(ish) process and can take a few minutes to complete depending on the size of the model and the number of line styles.

    Observations:

    - when deleting a line style in Revit, the associated graphics style will be deleted as well. Which leads to a minimum of 2 deleted elements per line style.
    - a custom element deleted modifier is used to check if the deleted element count is 2 and if the second deleted element is the associated graphics style.
    - if that is the case and no modified elements are listed, the deleted element count is reduced to 1 and the line style is considered as deleted.

    
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param progress_callback: Callback to report progress.
    :type progress_callback: callable
    :param debug: Debug mode.
    :type debug: bool
    :return: Result class instance.

        - .status True if unused line styles where deleted or nothing needed to be deleted. Otherwise False.
        - .message will contain deletion status.

    """

    return_value = Result()

    try:
        # set up delete modifier instance
        mod_delete = LineStylePurgeModifier(doc)

        # purge unused line styles
        purge_result = purge_unused_elements(
            doc=doc,
            element_id_getter=get_all_line_style_ids,
            deleted_elements_modifier=mod_delete, # reduces the count of deleted elements to the line style only by removing the associated graphics style
            modified_elements_modifier=None,
            progress_callback=progress_callback,
            debug=debug,
        )
        
        # get the debug log if debug mode is enabled
        if debug:
            return_value.append_message("\n".join(mod_delete.debug_log))

        return_value.update(purge_result)
    except Exception as e:
        return_value.update_sep(
            False, "Error purging line styles by delete: {}".format(str(e))
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data storage to Category override object helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

# used to convert integer to byte value
import System

from Autodesk.Revit.DB import Color, ElementId, LinePatternElement, OverrideGraphicSettings

from duHast.Revit.Common.Utility.revit_to_data_conversion import (
    VIEW_DETAIL_LEVEL_NAME_MAPPING_REVERSE,VIEW_DETAIL_LEVEL_NAME_MAPPING,
)

from duHast.Revit.LinePattern.line_patterns import build_patterns_dictionary_by_name
from duHast.Revit.LinePattern.fill_patterns import pattern_ids_by_name
from duHast.Revit.Common.Objects.Data.pattern_settings_base import PatternSettingBase

def get_colour(data_colour):
    """
    Creates a Revit Color object instance from a ColourBase object

    :param data_colour: A ColourBase instance.
    :type data_colour: :class: ColourBase
    :return: None if ColourBase values are -1, or larger then 255, otherwise a Color instance.
    :rtype: Autodesk.Revit.DB.Color
    """

    if data_colour.red == -1 and data_colour.green == -1 and data_colour.blue == -1:
        # reset to no colour value override
        return Color.InvalidColorValue
    else:
        try:
            return Color(
                System.Convert.ToByte(data_colour.red),
                System.Convert.ToByte(data_colour.green),
                System.Convert.ToByte(data_colour.blue),
            )
        except:
            return None


def get_fill_pattern_id(doc, data_pattern):
    """
    Returns the first fill pattern id in model matching name provided, or None if none was found.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param data_pattern: A pattern settings base object instance
    :type data_pattern: :class:`.PatternSettingBase`

    :return: The pattern id if a match eas found or Invalid Element Id  (-1)
    :rtype: Autodesk.Revit.DB.ElementId
    """

    all_patterns = pattern_ids_by_name(doc)
    if data_pattern.name in all_patterns:
        return all_patterns[data_pattern.name][0]
    else:
        return ElementId.InvalidElementId


def get_line_pattern_id(doc, data_pattern):
    """
    Returns the first line pattern id in model matching name provided, or None if none was found.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param data_pattern: A pattern settings base object instance
    :type data_pattern: :class:`.PatternSettingBase`

    :return: The pattern id if a match eas found or Invalid Element Id  (-1)
    :rtype: Autodesk.Revit.DB.ElementId
    """

    all_patterns = build_patterns_dictionary_by_name(doc)
    if data_pattern.name in all_patterns:
        return all_patterns[data_pattern.name][0]
    elif data_pattern.name == PatternSettingBase.SOLID_PATTERN:
        return LinePatternElement.GetSolidPatternId()
    else:
        return ElementId.InvalidElementId


def convert_to_revit_graphic_override(doc, data_override, is_filter_override=False):
    """
    Creates a Revit OverrideGraphicSettings instance from settings stored in an OverrideByCategory object.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param data_override: An OverrideByCategory object
    :type data_override: :class:`.OverrideByCategory`
    :param is_filter_override: Flag indicating an override for a filter (True), defaults to False
    :type is_filter_override: bool, optional
    :return: An OverrideGraphicSettings instance
    :rtype: Autodesk.Revit.DB.OverrideGraphicSettings
    """

    # set up a new revit api override settings instance
    revit_override_setting = OverrideGraphicSettings()
    
    # set values based on data override if one is present
    if data_override.are_overrides_present:
        # Sets the override color of the background pattern of cut faces.
        revit_override_setting.SetCutBackgroundPatternColor(
            get_colour(data_override.override_cut.pattern_background.colour)
        )
        # Sets the ElementId of the cut face background pattern override. The fill pattern must be a drafting pattern.
        # A value of InvalidElementId means no override is set.
        revit_override_setting.SetCutBackgroundPatternId(
            get_fill_pattern_id(
                doc, data_override.override_cut.pattern_background.fill_pattern_setting
            )
        )
        # Sets the visibility of the cut face background fill pattern.
        revit_override_setting.SetCutBackgroundPatternVisible(
            (data_override.override_cut.pattern_background.is_visible)
        )
        # Sets the override color of the foreground pattern of cut faces.
        revit_override_setting.SetCutForegroundPatternColor(
            get_colour(data_override.override_cut.pattern_foreground.colour)
        )
        # Sets the ElementId of the cut face foreground pattern override. The fill pattern must be a drafting pattern.
        # A value of InvalidElementId means no override is set.
        revit_override_setting.SetCutForegroundPatternId(
            get_fill_pattern_id(
                doc, data_override.override_cut.pattern_foreground.fill_pattern_setting
            )
        )
        # Sets the visibility of the cut face foreground fill pattern.
        revit_override_setting.SetCutForegroundPatternVisible(
            (data_override.override_cut.pattern_foreground.is_visible)
        )
        # Sets the cut surface line color.
        revit_override_setting.SetCutLineColor(
            get_colour(data_override.override_cut.line_cut.colour)
        )
        # Sets the ElementId of the cut surface line pattern.
        revit_override_setting.SetCutLinePatternId(
            get_line_pattern_id(
                doc, data_override.override_cut.line_cut.line_pattern_settings
            )
        )
        # Sets the cut surface line weight.
        revit_override_setting.SetCutLineWeight(
            data_override.override_cut.line_cut.weight
        )

        # detail override only exists on category overrides
        if is_filter_override == False:
            # Sets the detail level if required
            if(data_override.detail_level != -1):
                revit_override_setting.SetDetailLevel(
                    VIEW_DETAIL_LEVEL_NAME_MAPPING_REVERSE[data_override.detail_level]
                )
        # Sets the halftone value.
        revit_override_setting.SetHalftone(data_override.halftone)

        # Sets the projection surface line color.
        revit_override_setting.SetProjectionLineColor(
            get_colour(data_override.override_projection.line_projection.colour)
        )
        # Sets the ElementId of the projection surface line pattern.
        revit_override_setting.SetProjectionLinePatternId(
            get_line_pattern_id(
                doc,
                data_override.override_projection.line_projection.line_pattern_settings,
            )
        )
        # Sets the projection surface line weight.
        revit_override_setting.SetProjectionLineWeight(
            data_override.override_projection.line_projection.weight
        )
        # Sets the override color of the surface background pattern.
        revit_override_setting.SetSurfaceBackgroundPatternColor(
            get_colour(data_override.override_projection.pattern_background.colour)
        )
        # Sets the ElementId of the surface background pattern override. The fill pattern must be a drafting pattern.
        # A value of InvalidElementId means no override is set.
        revit_override_setting.SetSurfaceBackgroundPatternId(
            get_fill_pattern_id(
                doc,
                data_override.override_projection.pattern_background.fill_pattern_setting,
            )
        )
        # Sets the visibility of the surface background fill pattern.
        revit_override_setting.SetSurfaceBackgroundPatternVisible(
            data_override.override_projection.pattern_background.is_visible
        )
        # Sets the override color of the surface foreground pattern.
        revit_override_setting.SetSurfaceForegroundPatternColor(
            get_colour(data_override.override_projection.pattern_foreground.colour)
        )
        # Sets the ElementId of the surface foreground pattern override. The fill pattern must be a drafting pattern.
        # A value of InvalidElementId means no override is set.
        revit_override_setting.SetSurfaceForegroundPatternId(
            get_fill_pattern_id(
                doc,
                data_override.override_projection.pattern_foreground.fill_pattern_setting,
            )
        )
        # Sets the visibility of the surface foreground fill pattern.
        revit_override_setting.SetSurfaceForegroundPatternVisible(
            data_override.override_projection.pattern_foreground.is_visible
        )
        # Sets the projection surface transparency
        revit_override_setting.SetSurfaceTransparency(data_override.transparency)

    return revit_override_setting



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Interface for family processing class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
import Autodesk.Revit.DB as rdb
import json
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities.Objects import result as res
from duHast.Utilities.Objects import base


class IFamilyProcessor(base.Base):
    def __init__(
        self,
        data_type="not declared",
        pre_actions=None,
        post_actions=None,
        string_report_headers=[],
        **kwargs
    ):

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(IFamilyProcessor, self).__init__(**kwargs)

        self.data = []
        self.data_type = data_type
        self.string_report_headers = string_report_headers
        self.pre_actions = pre_actions
        self.post_actions = post_actions

    # -------------------------------------- utility ----------------------

    def _update_data(
        self,
        processor,
        identify_by_this_property_name,
        identify_by_this_property_value,
        update_by_property_name,
        updated_to_this_property_value,
    ):
        update_status = processor.update_Data(
            identify_by_this_property_name,
            identify_by_this_property_value,
            update_by_property_name,
            updated_to_this_property_value,
        )
        return update_status

    def _find_root_family_processor(self):
        """
        Finds the processor instance which processed the root family.

        :return: Processor instance
        :rtype: IFamilyProcessor
        """

        for processor in self.data:
            for d in processor.get_data():
                if " :: " not in d[IFamData.ROOT]:
                    return processor

    def _find_root_family_data(self):
        """
        Returns all data from root families (top most in tree) from all processor instances.

        :param data: List of dictionaries.
        :type data: [{}]

        :return: List of dictionaries.
        :rtype: [{}]
        """

        family_data = []
        for processor in self.data:
            for d in processor.get_data():
                if " :: " not in d[IFamData.ROOT]:
                    family_data.append(d)
        return family_data

    def _find_nested_families_data(self):
        """
        Returns all data from nested families from each processor instances.

        :param data: List of dictionaries.
        :type data: [{}]

        :return: List of dictionaries.
        :rtype: [{}]
        """

        nested_family_data = []
        for processor in self.data:
            for d in processor.get_data():
                if " :: " in d[IFamData.ROOT]:
                    nested_family_data.append(d)
        return nested_family_data

    def _fix_data_types(self, flattened_dic):
        """
        Replace any ElementId and Byte values with int or string respectively to have JSON working ok.
        Any other type of values are not changed.

        :param flattened_dic: Dictionary of which values are to be converted.
        :type flattened_dic: {}
        :return: Dictionary with converted values.
        :rtype: {}
        """

        dic = {}
        for key in flattened_dic:
            if type(flattened_dic[key]) is rdb.ElementId:
                dic[key] = flattened_dic[key].IntegerValue
            elif type(flattened_dic[key]) is System.Byte:
                dic[key] = str(flattened_dic[key])
            else:
                dic[key] = flattened_dic[key]
        return dic

    # -------------------------------------- pre process actions ----------------------

    def pre_process_actions(self, doc):
        """
        Actions any pre processing before family data will be collected.

        :param doc: The family document.
        :type doc: Autodesk.Revit.DB.Document

        :return: _description_
        :rtype: _type_
        """

        return_value = res.Result()
        if self.pre_actions != None:
            for pre_action in self.pre_actions:
                result_action = pre_action(doc)
                return_value.update(result_action)
        return return_value

    # -------------------------------------- process actions ----------------------

    def process(self, doc, root_path, root_category_path):
        """
        Gather data on the root family and any nested families

        :param doc: The family document. 
        :type doc: Autodesk.Revit.DB.Document

        :param root_path: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_path: str
        :param root_category_path: The path of the nested family category in a tree: rootFamilyCategory::nestedFamilyOneCategory::nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type root_category_path: str
        """

        pass

    # -------------------------------------- post process actions ----------------------

    def post_process_actions(self, doc):
        """
        Actions any post processing after family data has been collected.

        :param doc: The family document.
        :type doc: Autodesk.Revit.DB.Document
        """

        return_value = res.Result()
        if self.post_actions != None:
            for post_action in self.post_actions:
                result_action = post_action(doc)
                return_value.update(result_action)
        return return_value

    # -------------------------------------- get data ----------------------

    def get_data(self):
        """
        Returns list of flattened dictionaries. One dictionary for each document processed.

        :return: List of dictionaries.
        :rtype: [{}]
        """

        data_out = []
        for data in self.data:
            for d in data.get_data():
                data_out.append(d)
        return data_out

    def get_data_json(self):
        """
        Returns data objects as JSON formatted strings.

        :return: JSON formatted string.
        :rtype: str
        """

        out_value = ""
        flattened_data = self.get_data()
        for d in flattened_data:
            d_fixed_types = self._fix_data_types(d)
            json_object = json.dumps(dict(d_fixed_types))
            out_value = out_value + "\n" + json_object
        return out_value

    def get_data_string_list(self):
        """
        Returns data objects as list of strings in order of headers list of this class.

        - Strings are UTF 8 encoded
        - Unknown header values are marked as 'null'

        :return: list of string.
        :rtype: [str]
        """
        out_value = []
        flattened_data = self.get_data()
        for d in flattened_data:
            row = []
            for header_key in self.string_report_headers:
                if header_key in d:
                    value = None
                    if type(d[header_key]) == str:
                        # make sure string is utf-8 encoded
                        value = d[header_key].encode("utf-8", "ignore")
                    else:
                        value = str(d[header_key])
                    row.append(value)
                else:
                    row.append("null")
            out_value.append(row)
        return out_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit grids helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

import System

# import common library modules
from duHast.Revit.Common import common as com
from duHast.Revit.Common import parameter_get_utils as rParaGet

# import Autodesk
import Autodesk.Revit.DB as rdb


def get_grids_in_model(doc):
    """
    Get all grids in model

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector with all grids in model.
    :rtype: _type_
    """

    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.Grid)
    return collector


# --------------------------------------------- utility functions ------------------


def get_all_grid_heads_by_category(doc):
    """
    Gets all grid head types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector with grid head types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_GridHeads)
        .WhereElementIsElementType()
    )
    return collector


def get_all_grid_types_by_category(doc):
    """
    Gets all grid types in the model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector with grid types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_Grids)
        .WhereElementIsElementType()
    )
    return collector


def get_all_grid_type_ids_by_category(doc):
    """
    Gets all grid types ids in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector with grid type ids
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = get_all_grid_types_by_category(doc)
    ids = com.get_ids_from_element_collector(collector)
    return ids


def get_grid_type_names(doc, g):
    """
    Gets all valid grid types, based on a past in grid, available in model.

    Uses grid.GetValidTypes() to get the grid types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param g: A grid
    :type g: Autodesk.Revit.DB.Grid
    :return: A nested set of lists containing grid type id and grid type name
    :rtype: list of lists [[GridTypeId as Revit ElementId, grid type name as string],[...]]
    """

    valid_grid_types = []
    valid_grid_type_ids = g.GetValidTypes()
    for valid_grid_type_id in valid_grid_type_ids:
        grid_data = []
        grid_type_t = doc.GetElement(valid_grid_type_id)
        grid_data.append(valid_grid_type_id)
        grid_data.append(rdb.Element.Name.GetValue(grid_type_t))
        valid_grid_types.append(grid_data)
    return valid_grid_types


def get_grid_type_name(doc, g):
    """
    Gets the grid type name of a grid.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :return: The grid type name.
    :rtype: str
    """

    value = "unknown"
    grid_type_T = doc.GetElement(g.GetTypeId())
    value = rdb.Element.Name.GetValue(grid_type_T)
    return value


def get_grid_type_id_by_name(doc, grid_type_name):
    """
    Gets the grid type Id based on it's name, if no match found it returns the Revit Invalid Element Id

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param grid_type_name: The grid type name.
    :type grid_type_name: str
    :return: The grids type Id or if not match is found Autodesk.Revit.DB.ElementId.InvalidElementId
    :rtype: Autodesk.Revit.DB.ElementId
    """

    id = rdb.ElementId.InvalidElementId
    grids = rdb.FilteredElementCollector(doc).OfClass(rdb.Grid).ToList()
    if len(grids) > 0:
        g = grids[0]
        valid_grid_type_ids = g.GetValidTypes()
        for grid_typ_id in valid_grid_type_ids:
            g_type_name = rdb.Element.Name.GetValue(doc.GetElement(grid_typ_id))
            if g_type_name == grid_type_name:
                id = grid_typ_id
                break
    return id


def grid_check_parameter_value(g, para_name, para_condition, condition_value):
    """
    Returns true if a given parameter on a grid has a value meeting the parameter condition.

    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :param para_name: A parameter Name.
    :type para_name: str
    :param para_condition: A function evaluating the parameter value. First argument is the value to be checked against. Second argument is the actual parameter value.
    :type para_condition: func(arg1,arg2)
    :param condition_value: The value to be checked against.
    :type condition_value: var
    :return: True if parameter value is evaluated to True otherwise False.
    :rtype: bool
    """

    rule_match = False
    p_value = rParaGet.get_parameter_value_by_name(g, para_name)
    if p_value != None:
        rule_match = rParaGet.check_parameter_value(g, para_condition, condition_value)
    return rule_match


def get_max_extent_as_string(g):
    """
    Gets the maximum extent of a grid.

    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :return: A string in format [maxX,maxY,maxZ]
    :rtype: str
    """

    ex = g.GetExtents()
    max = "[{}]".format(
        ",".join(
            [str(ex.MaximumPoint.X), str(ex.MaximumPoint.Y), str(ex.MaximumPoint.Z)]
        )
    )
    return max


def get_min_extent_as_string(g):
    """
    Gets the minimum extent of a grid.

    :param g: A grid.
    :type g: Autodesk.Revit.DB.Grid
    :return: A string in format [minX,minY,minZ]
    :rtype: str
    """

    ex = g.GetExtents()
    min = "[{}]".format(
        ",".join(
            [str(ex.MinimumPoint.X), str(ex.MinimumPoint.Y), str(ex.MinimumPoint.Z)]
        )
    )
    return min


def get_all_grid_head_family_type_ids(doc):
    """
    This will return all ids grid head family types in the model
    """

    ids = []
    filter = rdb.ElementCategoryFilter(rdb.BuiltInCategory.OST_GridHeads)
    col = (
        rdb.FilteredElementCollector(doc).OfClass(rdb.FamilySymbol).WherePasses(filter)
    )
    ids = com.get_ids_from_element_collector(col)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit levels and level heads. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Family import purge_unused_family_types as rFamPurge
from duHast.Revit.Common import (
    parameter_get_utils as rParaGet,
    purge_utils as rPurgeUtils,
)
from duHast.Revit.Levels import levels as rLevel


def get_unused_level_types_for_purge(doc):
    """
    Gets all ids of unused level types in the model.
    Unused: not one instance per level type is placed in the model.
    This method can be used to safely delete unused level types from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of level type ids.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    return rPurgeUtils.get_used_unused_type_ids(
        doc, rLevel.get_all_level_type_ids_by_category, 0, 6
    )


def get_unused_level_head_families(doc):
    """
    Gets all ids of unused family symbols (types) of level head families.
    Unused: not one instance per symbol is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of symbol ids.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    usedTypes = rPurgeUtils.get_used_unused_type_ids(
        doc, rLevel.get_all_level_type_ids_by_category, 1, 6
    )
    headsInUseIds = []
    # get family symbol in use at level as symbol
    for lId in usedTypes:
        type = doc.GetElement(lId)
        id = rParaGet.get_built_in_parameter_value(
            type, rdb.BuiltInParameter.LEVEL_HEAD_TAG
        )
        if id != None and id not in headsInUseIds:
            headsInUseIds.append(id)
    # get all level head symbols available
    allSymbolsInModel = rLevel.get_all_level_heads_by_category(doc)
    unusedSymbolIds = []
    # filter out unused level head symbols and add to list to be returned
    for levelSymbolInModel in allSymbolsInModel:
        if levelSymbolInModel.Id not in headsInUseIds:
            unusedSymbolIds.append(levelSymbolInModel.Id)
    return unusedSymbolIds


def get_unused_level_head_families_for_purge(doc):
    """
    Gets ids of all unused level head symbols and families.
    Unused: not one instance per level symbol is placed in the model.
    This method can be used to safely delete unused level symbols or families from the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of symbol and or family ids.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    return rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_level_head_families
    )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Views purge unused utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import common as com
from duHast.Revit.Family import family_utils as rFamUPurge
from duHast.Revit.Views import referencing as rViewRef
from duHast.Revit.Views import views as rView
from duHast.Revit.Family import purge_unused_family_types as rFamPurge


# view reference purging


def get_unused_continuation_marker_type_ids_for_purge(doc):
    """
    Gets all unused view continuation type ids in model for purge.
    This method can be used to safely delete all unused view continuation marker types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    all_available_type_ids = rViewRef.get_all_view_continuation_type_ids(doc)
    all_used_type_ids = rViewRef.get_used_view_continuation_type_ids(doc)
    for a_id in all_available_type_ids:
        if a_id not in all_used_type_ids:
            ids.append(a_id)
    return ids


def is_nested_family_symbol(doc, id, nested_family_names):
    """
    Returns true if symbol belongs to family in list past in.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param id: The element id of a symbol.
    :type id: Autodesk.Revit.DB.ElementId
    :param nested_family_names: list of family names know to be nested families.
    :type nested_family_names: list str
    :return: True if family name derived from symbol is in list past in, otherwise False.
    :rtype: bool
    """

    flag = False
    fam_symbol = doc.GetElement(id)
    fam = fam_symbol.Family
    if fam.Name in nested_family_names:
        flag = True
    return flag


def get_unused_view_ref_and_continuation_marker_symbol_ids(doc):
    """
    Gets the ids of all view reference symbols(types) and view continuation symbols (types) not used in the model.
    Not used: These symbols are not used in any view reference types, or nested in any symbols used in view reference types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    # compare used vs available in view ref types
    # whatever is marked as unused: check for any instances in the model...placed on legends!
    available_ids = rViewRef.get_all_view_reference_symbol_ids(
        doc
    )  # check: does this really return all continuation marker types??
    used_ids = rViewRef.get_used_view_reference_and_continuation_marker_symbol_ids(doc)
    # elevation marker families might use nested families...check!
    nested_family_names = rViewRef.get_nested_family_marker_names(doc, used_ids)
    check_ids = []
    for a_id in available_ids:
        if a_id not in used_ids:
            check_ids.append(a_id)
    # check for any instances
    for id in check_ids:
        instances = rFamUPurge.get_family_instances_by_symbol_type_id(doc, id).ToList()
        if len(instances) == 0:
            if is_nested_family_symbol(doc, id, nested_family_names) == False:
                ids.append(id)
    return ids


def get_unused_view_ref_and_continuation_marker_families_for_purge(doc):
    """
    Gets the ids of all view reference symbols(types) ids and or family ids not used in the model for purging.
    This method can be used to safely delete all unused view reference and continuation marker family symbols\
        or families.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    return rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_view_ref_and_continuation_marker_symbol_ids
    )


def get_unused_view_reference_type_ids_for_purge(doc):
    """
    Gets all unused view references type ids in model for purge.
    This method can be used to safely delete all unused view reference types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    all_available_type_ids = rViewRef.get_all_view_reference_type_id_data(doc)
    all_used_type_ids = rViewRef.get_used_view_reference_type_id_data(doc)
    for key, value in all_available_type_ids.items():
        if all_used_type_ids.has_key(key):
            for available_type_id in all_available_type_ids[key]:
                if available_type_id not in all_used_type_ids[key]:
                    ids.append(available_type_id)
        else:
            # add all types under this key to be purge list...might need to check whether I need to leave one behind...
            if len(all_available_type_ids[key]) > 0:
                ids = ids + all_available_type_ids[key]
    return ids


# view types purging


def get_used_view_type_ids(doc):
    """
    Returns all view family types in use in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: ids of view family types in use
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    view_type_ids_used = []
    col = rdb.FilteredElementCollector(doc).OfClass(rdb.View)
    for v in col:
        # filter out browser organization and other views which cant be deleted
        if (
            v.IsTemplate == False
            and v.ViewType != rdb.ViewType.SystemBrowser
            and v.ViewType != rdb.ViewType.ProjectBrowser
            and v.ViewType != rdb.ViewType.Undefined
            and v.ViewType != rdb.ViewType.Internal
            and v.ViewType != rdb.ViewType.DrawingSheet
        ):
            if v.GetTypeId() not in view_type_ids_used:
                view_type_ids_used.append(v.GetTypeId())
    return view_type_ids_used


def get_unused_view_type_ids(doc):
    """
    Returns all unused view family types in the model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: ids of view family types not in use
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    filtered_unused_view_type_ids = com.get_unused_type_ids_in_model(
        doc, rView.get_view_types, get_used_view_type_ids
    )
    return filtered_unused_view_type_ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit filter overrides in views. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import clr
import System

# import common library modules
from duHast.Revit.Views.Objects.filter_override_storage import RevitFilterOverride
from duHast.Revit.Views.filters import get_all_filters
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Views.filters import VIEW_TYPE_WHICH_CAN_HAVE_FILTERS

# from duHast.APISamples.Common import RevitElementParameterGetUtils as rParaGet
from duHast.Utilities.Objects import result as res

# import Autodesk
from Autodesk.Revit.DB import Transaction

# --------------------------- view filters --------------------------------


def get_filters_from_model(doc):
    """
    Retrieves all filters in a Revit model and stores them in a dictionary.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary containing the filters in the model as RevitFilterOverride objects. The keys of the dictionary are the filter names.
    :rtype: {str: :class:`.RevitFilterOverride`}
    """
    filters_dic = {}

    # get all filters in the model
    filters_in_model = get_all_filters(doc)
    for f in filters_in_model:
        filters_dic["{}".format(f.Name)] = RevitFilterOverride(
            filter_name=f.Name,
            filter=f,
            filter_id=f.Id,
            filter_override=None,
            is_filter_enabled=True,
            is_filter_visible=False,
        )

    return filters_dic


def update_filter_override_from_view(view, filter_storage_instance):
    """
    Populates the filter_override and is_filter_visible and is_filter_enabled field of a 'RevitFilterOverride' instance based on the view past in.

    :param view: The view from which the filter override is to be used.
    :type view: Autodesk.Revit.DB.View
    :param filter_storage_instance: An instances of revit filter storage containing no view specific information.
    :type filter_storage_instance: :class:`.RevitFilterOverride`

    :return: An updated instances of revit filter storage now containing the override, filter visibility and filter enabled flags.
    :rtype: :class:`.RevitFilterOverride`
    """

    filter_storage_instance.filter_override = view.GetFilterOverrides(
        filter_storage_instance.filter_id
    )
    filter_storage_instance.is_filter_visible = view.GetFilterVisibility(
        filter_storage_instance.filter_id
    )
    filter_storage_instance.is_filter_enabled = view.GetIsFilterEnabled(
        filter_storage_instance.filter_id
    )

    return filter_storage_instance


def get_filter_overrides_from_view(view, filter_storage_instances):
    """
    Populates the category_override and is_category_hidden fields of a list of 'category_storage' instances based on the view past in.

    :param view: The view from which the category override is to be used.
    :type view: Autodesk.Revit.DB.View
    :param filter_storage_instances: A list of instances of class 'RevitFilterOverride'
    :type filter_storage_instances: [:class:`.RevitFilterOverride`]

    :return: A list of instances of class 'RevitFilterOverride' where the override, filter visibility and filter enabled flags have been updated dependent on the view.
    :rtype: [:class:`.RevitFilterOverride`]
    """

    updated_filter_storage_instances = []
    for filter_storage_instance in filter_storage_instances:
        updated_filter_instance = update_filter_override_from_view(
            view, filter_storage_instance
        )
        updated_filter_storage_instances.append(updated_filter_instance)
    return updated_filter_storage_instances


def apply_filter_override_to_view(doc, view, filter_storage_instances):
    """
    Applies filter override(s) to a single view

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view on which the category override is to be used.
    :type view: Autodesk.Revit.DB.View
    :param filter_storage_instances: A list of storage class instances.
    :type filter_storage_instances: [:class:`.RevitFilterOverride`]

    :return:
        Result class instance.

        - Apply override status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain message 'Successfully set category override...' for each override applied.
        - result.result will be an empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result will be an empty list.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for filter_storage_instance in filter_storage_instances:

        def action():
            action_return_value = res.Result()
            try:
                view.SetFilterOverrides(
                    filter_storage_instance.filter_id,
                    filter_storage_instance.revit_override,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully set filter override {} in view {} ".format(
                        filter_storage_instance.filter_name,
                        view.Name,
                    ),
                )
                # filter visibility
                view.SetFilterVisibility(
                    filter_storage_instance.filter_id,
                    filter_storage_instance.is_filter_visible,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully set filter visibility status {} in view {} to: {}".format(
                        filter_storage_instance.filter_name,
                        view.Name,
                        filter_storage_instance.is_filter_visible,
                    ),
                )

                # filter enabled
                view.SetIsFilterEnabled(
                    filter_storage_instance.filter_id,
                    filter_storage_instance.is_filter_enabled,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully set filter enabled status {} in view {} to: {}".format(
                        filter_storage_instance.filter_name,
                        view.Name,
                        filter_storage_instance.is_filter_enabled,
                    ),
                )

            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set filter override : {} to view: {} with exception: {}".format(
                        filter_storage_instance.filter_name,
                        view.Name,
                        e,
                    ),
                )
            return action_return_value

        transaction = Transaction(
            doc,
            "Updating filter override {}".format(
                filter_storage_instance.filter_name,
            ),
        )
        update_category_override = in_transaction(transaction, action)
        return_value.update(update_category_override)

    return return_value


def view_has_filter(view, filter_to_check_id):
    """
    Checks if a specific filter is applied to a given view.

    :param view: The view object to check for applied filters.
    :type view: Autodesk.Revit.DB.View
    :param filter_to_check_id: The ID of the filter to check if it is applied.
    :type filter_to_check_id: Autodesk.Revit.DB.ElementId

    :return: A boolean value indicating whether the specified filter is applied to the view.
    :rtype: bool
    """
    filter_is_applied = False
    filters_applied_to_view_as_ids = view.GetFilters()
    if filter_to_check_id in filters_applied_to_view_as_ids:
        filter_is_applied = True
    return filter_is_applied


def add_filter_to_view(doc, filter, view):
    """
    Adds a filter to a view in Autodesk Revit.

    :param doc: The current model document in Autodesk Revit.
    :type doc: Autodesk.Revit.DB.Document
    :param filter: The filter to be added to the view.
    :type filter: Autodesk.Revit.DB.ParameterFilterElement
    :param view: The view to which the filter will be added.
    :type view: Autodesk.Revit.DB.View

    :return: An instance of the Result class that contains the result of the filter addition. The result attribute is True if the filter was added successfully, otherwise False. The message attribute contains additional information about the result.
    :rtype: :class:`.Result`
    """
    return_value = res.Result()

    def action():
        action_return_value = res.Result()
        try:
            view.AddFilter(filter.id)
            return_value.append_message(
                "Filter: {} added to view: {} successfully.".format(
                    filter.Name, view.Name
                )
            )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to apply filter: {} to view: {} with exception: {}".format(
                    filter.Name,
                    view.Name,
                    e,
                ),
            )
        return action_return_value

    transaction = Transaction(
        doc,
        "Adding filter: {} to view: {}".format(
            filter.Name,
            view.Name,
        ),
    )
    add_filter_status = in_transaction(transaction, action)
    return_value.update(add_filter_status)
    return return_value


def remove_filter_from_view(doc, filter, view):
    """
    Removes a filter from a view.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param filter: The filter to be removed from the view.
    :type filter: Autodesk.Revit.DB.ParameterFilterElement
    :param view: The view (can be a view template too!)
    :type view: Autodesk.Revit.DB.View
    :return:
        Result class instance.

        - .result True if filter was removed successfully or if filter was not applied or if view does not support filters, otherwise False
        - .message will contain deletion status.
        - . result (empty list)

        on exception:

        - .result Will be False
        - .message will contain exception message.
        - . result (empty list)

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    if view.ViewType in VIEW_TYPE_WHICH_CAN_HAVE_FILTERS:
        filters_applied = view.GetFilters()
        if filter.Id in filters_applied:

            def action():
                action_return_value = res.Result()
                try:
                    view.RemoveFilter(filter.Id)
                    action_return_value.update_sep(
                        True,
                        "Remove filter: {} from template: {}".format(
                            filter.Name, view.Name
                        ),
                    )
                except Exception as e:
                    action_return_value.update_sep(
                        False,
                        "Failed to remove filter: {} from template: {} with exception: {}".format(
                            filter.Name, view.Name, e
                        ),
                    )
                return action_return_value

            transaction = Transaction(doc, "Removing filter: {}".format(filter.Name))
            # execute the transaction
            return_value = in_transaction(transaction, action)
        else:
            return_value.update_sep(
                True,
                "Filter: {} is not applied to view template: {}".format(
                    filter.Name, view.Name
                ),
            )
    else:
        return_value.update_sep(
            True,
            "View template: {} if of type: {} which does not support filters.".format(
                view.Name, view.ViewType
            ),
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class used to implement deleted and modified element counts when using purge by delete.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Adjust deleted and modified elements if necessary:
this might be required where an element is presented through 2 or more elements in the revit api:
e.g. a line style is represented to a line style and a graphics style
this function will only purge elements which result in only 1 element  deleted and no other element modified
hence a delete modifier should check the deleted elements and if appropriate return only one element to be deleted
same applies to modified elements: a custom modifier should return 0 elements if appropriate in order for the element to be purged
   
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base


class ModifierBase(base.Base):
    def __init__(self):
        """
        Class constructor.

        """

        super(ModifierBase, self).__init__()

    def modify_deleted(self, doc, deleted):
        """
        Base implementation to modify the deleted element count.

        Returns deleted element list unchanged

        Args:
            deleted: The deleted element count

        """

        if isinstance(deleted, list) == False:
            raise TypeError("deleted must be a list")

        return deleted

    def modify_modified(self, doc, modified):
        """
        Base implementation to modify the modified element count.

        Returns modified element list unchanged

        Args:
            modified: The modified element count

        """

        if isinstance(modified, list) == False:
            raise TypeError("modified must be a list")

        return modified



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of functions around Revit annotation objects.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
from duHast.Revit.Annotation.arrow_heads import get_arrow_head_ids_from_type, ARROWHEAD_PARAS_TEXT

# import Autodesk
from Autodesk.Revit.DB import AnnotationSymbolType, AreaTagType, FamilySymbol, FilteredElementCollector
from Autodesk.Revit.DB.Architecture import RoomTagType

# ----------------------------------------------


def get_all_anno_symbol_types(doc):
    """
    Gets all annotation symbol types, area tag types, room tag types in the model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: list of types
    :rtype: list
    """

    types = []
    col = FilteredElementCollector(doc).OfClass(FamilySymbol)
    for c in col:
        if (
            c.GetType() == AnnotationSymbolType
            or c.GetType == AreaTagType
            or c.GetType() == RoomTagType
        ):
            types.append(c)
    return types


def get_anno_symbol_arrow_head_ids(doc):
    """
    Gets all arrow head ids used in annotation symbol types, area tag types, room tag types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing arrow head symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    used_ids = get_arrow_head_ids_from_type(
        doc, get_all_anno_symbol_types, ARROWHEAD_PARAS_TEXT
    )
    return used_ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a function to bind a shared parameter to a category.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
based on building coder article:
https://thebuildingcoder.typepad.com/blog/2012/04/adding-a-category-to-a-shared-parameter-binding.html

"""


import Autodesk.Revit.DB as rdb

# custom result class
from duHast.Utilities.Objects import result as res

# import InTransaction from common module
from duHast.Revit.Common import transaction as rTran


def load_shared_parameter_file(doc, path=None):
    """
    Loads a shared parameter file.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param path: (Optional) Fully qualified file path to shared parameter text file.
    :type path: str

    :return: The opened shared parameter file.
    :rtype: Autodesk.Revit.DB.DefinitionFile
    """

    app = doc.Application
    if path:
        app.SharedParametersFilename = path

    return app.OpenSharedParameterFile()


def bind_shared_parameter(
    doc,
    category,
    parameter_name,
    group_name,
    parameter_type,
    is_visible,
    is_instance,
    parameter_grouping,
    shared_parameter_filepath,
):
    """
    Binds a shared parameter to a revit category.

    Refer building coder article referenced in header


    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param category: The built in category, to which the parameter will be bound.
    :type category: Autodesk.Revit.DB.BuiltInCategory
    :param parameter_name: The parameter name.
    :type parameter_name: str
    :param group_name: The group under which the parameter appears in shared parameter text file.
    :type group_name: str
    :param paramType: The parameter type. (Area, vs text vs... (deprecated in Revit 2022!)
    :type paramType: Autodesk.Revit.DB.ParameterType
    :param is_visible: Is parameter visible in UI to users.
    :type is_visible: bool
    :param is_instance: True parameter is an instance parameter, otherwise type parameter.
    :type is_instance: bool
    :param parameter_grouping: Where parameter appears in properties section in UI.
    :type parameter_grouping: str
    :param shared_parameter_filepath: Fully qualified file path to shared parameter text file.
    :type shared_parameter_filepath: str

    :return:
        Result class instance.

        - Parameter binding status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the shared parameter.

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        app = doc.Application

        # check if we are going to get something valid:
        if doc.Settings.Categories.get_Item(category) != None:
            # This is needed already here to
            # store old ones for re-inserting
            cat_set = app.Create.NewCategorySet()

            # Loop all Binding Definitions
            # IMPORTANT NOTE: Categories.Size is ALWAYS 1 !?
            # For multiple categories, there is really one
            # pair per each category, even though the
            # Definitions are the same...

            iter = doc.ParameterBindings.ForwardIterator()
            iter.Reset()
            while iter.MoveNext():
                if iter.Key != None:
                    definition = iter.Key
                    elem_bind = iter.Current
                    # check parameter name match
                    if parameter_name == definition.Name:
                        try:
                            cat = doc.Settings.Categories.get_Item(category)
                            if elem_bind.Categories.Contains(cat):
                                # check parameter type
                                if definition.ParameterType != parameter_type:
                                    return_value.status = False
                                    return_value.message = (
                                        "{}: wrong parameter type: {}".format(
                                            parameter_name, definition.ParameterType
                                        )
                                    )
                                    return return_value
                                # check binding type
                                if is_instance:
                                    if elem_bind.GetType() != rdb.InstanceBinding:
                                        return_value.status = False
                                        return_value.message = "{}: wrong binding type (looking for instance but got type)".format(
                                            parameter_name
                                        )
                                        return return_value
                                else:
                                    if elem_bind.GetType() != rdb.TypeBinding:
                                        return_value.status = False
                                        return_value.message = "{}: wrong binding type (looking for type but got instance)".format(
                                            parameter_name
                                        )
                                        return return_value

                                # Check Visibility - cannot (not exposed)
                                # If here, everything is fine,
                                # ie already defined correctly
                                return_value.message = "{}: Parameter already bound to category: {}".format(
                                    parameter_name, cat.Name
                                )
                                return return_value
                        except Exception as e:
                            return_value.append_message(
                                "{} : Failed to check parameter binding with exception: {}".format(
                                    parameter_name, e
                                )
                            )
                        # If here, no category match, hence must
                        # store "other" cats for re-inserting
                        else:
                            for cat_old in elem_bind.Categories:
                                cat_set.Insert(cat_old)

            # If here, there is no Binding Definition for
            # it, so make sure Param defined and then bind it!
            def_file = load_shared_parameter_file(doc, shared_parameter_filepath)
            def_group = def_file.Groups.get_Item(group_name)
            if def_group == None:
                def_group = def_file.Groups.Create(group_name)
            if def_group.Definitions.Contains(
                def_group.Definitions.Item[parameter_name]
            ):
                definition = def_group.Definitions.Item[parameter_name]
            else:
                opt = rdb.ExternalDefinitionCreationOptions(
                    parameter_name, parameter_type
                )
                opt.Visible = is_visible
                definition = def_group.Definitions.Create(opt)

            # get category from builtin category
            cat_object = doc.Settings.Categories.get_Item(category)
            cat_set.Insert(cat_object)

            bind = None
            if is_instance:
                bind = app.Create.NewInstanceBinding(cat_set)
            else:
                bind = app.Create.NewTypeBinding(cat_set)

            # There is another strange API "feature".
            # If param has EVER been bound in a project
            # (in above iter pairs or even if not there
            # but once deleted), Insert always fails!?
            # Must use .ReInsert in that case.
            # See also similar findings on this topic in:
            # http://thebuildingcoder.typepad.com/blog/2009/09/adding-a-category-to-a-parameter-binding.html
            # - the code-idiom below may be more generic:

            def action():
                action_return_value = res.Result()
                try:
                    if doc.ParameterBindings.Insert(
                        definition, bind, parameter_grouping
                    ):
                        action_return_value.message = (
                            "{} : parameter successfully bound to: {}".format(
                                parameter_name, cat_object.Name
                            )
                        )
                        return action_return_value
                    else:
                        if doc.ParameterBindings.ReInsert(
                            definition, bind, parameter_grouping
                        ):
                            action_return_value.message = (
                                "{} : parameter successfully bound to: {}".format(
                                    parameter_name, cat_object.Name
                                )
                            )
                            return action_return_value
                        else:
                            action_return_value.status = False
                            action_return_value.message = (
                                "{} : failed to bind parameter to: {}".format(
                                    parameter_name, cat_object.Name
                                )
                            )
                except Exception as e:
                    action_return_value.status = False
                    action_return_value.message = "{} : Failed to bind parameter to: {} with exception: {}".format(
                        parameter_name, cat_object.Name, e
                    )
                return action_return_value

            transaction = rdb.Transaction(doc, "Binding parameter")
            return_value = rTran.in_transaction(transaction, action)
        else:
            return_value.update_sep(
                False, "Failed to get category object for: {}".format(category)
            )
        return return_value

    except Exception as e:
        return_value.status = False
        return_value.message = (
            "{} : Failed to bind parameter with exception: {}".format(parameter_name, e)
        )
    return return_value


def add_shared_parameter_to_family(para, mgr, doc, def_file):
    """
    Adds a shared parameter definition to a family document.

    :param para: Tuple containing parameter info
    :type para: tuple (refer module RevitSharedParametersTuple)
    :param mgr: The family manager object
    :type mgr: Autodesk.Revit.DB.FamilyManager
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param def_file: The shared parameter definition file.
    :type def_file: _type_

    :return:
        Result class instance.

        - True if added successfully. False if an exception occurred.
        - result.message will contain the name of the shared parameter.
        - .result.result will contain the family parameter object.

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    found_para = False
    try:
        # loop through parameters and try to find matching one to be added from parameter file
        # loop through all definition groups
        for group in def_file.Groups:
            # loop through para's within definition group
            for def_para in group.Definitions:
                # check whether this is the parameter we are after
                if def_para.Name != para.name:
                    # jump to next parameter
                    continue

                # set up an action to add parameter
                def action():
                    action_return_value = res.Result()
                    try:
                        # add parameter depending on name, parameter group and isInstance
                        fam_para = mgr.AddParameter(
                            def_para, para.builtInParameterGroup, para.isInstance
                        )
                        action_return_value.message = (
                            para.name + " : parameter successfully added."
                        )
                        action_return_value.result.append(fam_para)
                    except Exception as e:
                        action_return_value.status = False
                        action_return_value.message = (
                            para.name
                            + " : Failed to add shared parameter: with exception: "
                            + str(e)
                        )
                    return action_return_value

                transaction = rdb.Transaction(doc, "Adding shared parameter")
                return_value = rTran.in_transaction(transaction, action)
                # set flag for parameter found
                found_para = True

            # check whether inner loop found matching parameter
            if found_para:
                # get out of outer loop
                break
    except Exception as e:
        return_value.status = False
        return_value.message = (
            para.name + " : Failed to add parameter to family with exception: " + str(e)
        )

    if found_para == False:
        return_value.status = False
        return_value.message = (
            para.name + " : No match for parameter found in shared parameter file."
        )

    return return_value


def bind_shared_parameters_to_new_category(
    rvt_doc, target_params, target_cat, target_param_grp, type_binding=False
):
    """
    Takes a list of shared parameter definitions and creates a binding to a new category. Will
    add category to an existing binding if one exists already
    :param rvt_doc: Revit document
    :type rvt_doc: Autodesk.Revit.DB.Document
    :param target_params: List of shared parameter definitions
    :type target_params: list[Autodesk.Revit.DB.ExternalDefinition]
    :param target_cat: Category to bind to
    :type target_cat: BuiltInCategory
    :param target_param_grp: Parameter Group new parameters will appear in
    :type target_param_grp: BuiltInParameterGroup
    :param type_binding: True if type binding, False if instance binding
    :type type_binding: bool
    :return: List of successful bindings, list of errors
    :rtype: tuple
    """
    new_cat_set = rvt_doc.Application.Create.NewCategorySet()
    new_cat_set.Insert(target_cat)
    if type_binding:
        new_binding = rvt_doc.Application.Create.NewTypeBinding(new_cat_set)
    else:
        new_binding = rvt_doc.Application.Create.NewInstanceBinding(new_cat_set)
    binding_map = rvt_doc.ParameterBindings
    target_cat_name = rdb.Category.GetCategory(rvt_doc, target_cat.Id).Name

    outlist = []
    errors = []

    t = rdb.Transaction(rvt_doc, "Bind new category to shared parameters")
    t.Start()

    for sp_def in target_params:
        ex_binding = binding_map.Item[sp_def]
        if ex_binding:
            ex_binding_cats = [cat.Name for cat in ex_binding.Categories]
            if target_cat_name in ex_binding_cats:
                outlist.append(
                    "{} already bound to {}".format(sp_def.Name, target_cat_name)
                )
                continue
            try:
                ex_binding.Categories.Insert(target_cat)
                rvt_doc.ParameterBindings.ReInsert(sp_def, ex_binding)
                outlist.append("Bound {} to {}".format(sp_def.Name, target_cat.Name))
            except:
                errors.append(
                    "Could add {} to existing binding for {}".format(
                        target_cat.Name, sp_def.Name
                    )
                )
        else:
            try:
                binding_map.Insert(sp_def, new_binding, target_param_grp)
                outlist.append("Bound {} to {}".format(sp_def.Name, target_cat.Name))
            except:
                errors.append(
                    "Could not add {} to new binding for {}".format(
                        target_cat.Name, sp_def.Name
                    )
                )

    t.Commit()

    return outlist, errors



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to building MEP ype dictionaries.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TODO: similar function exists in Walls module. Consider more generic function.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def build_type_dictionary(collector, dic):
    """
    Returns the dictionary past in with keys and or values added retrieved from collector past in.

    TODO: similar function exists in Walls module. Consider more generic function.

    :param collector: A filtered element collector containing railing type elements of family symbols
    :type collector: Autodesk.Revit.DB.FilteredElementCollector
    :param dic: dictionary containing key: railing type family name, value: list of ids
    :type dic: Dictionary {str:[Autodesk.Revit.DB.ElementId]}

    :return: A dictionary where key is the family name and values are ids of types belonging to that family.
    :rtype: Dictionary {str:[Autodesk.Revit.DB.ElementId]}
    """

    for c in collector:
        if dic.has_key(c.FamilyName):
            # todo : check WallKind Enum???
            if c.Id not in dic[c.FamilyName]:
                dic[c.FamilyName].append(c.Id)
        else:
            dic[c.FamilyName] = [c.Id]
    return dic


def sort_types_by_family_name(doc, type_getter):
    """
    Returns a dictionary where key is the family name and values are ids of types belonging to that family.
    TODO: similar function exists in Walls module. Consider more generic function.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_getter: function returning element types
    :type type_getter: func(doc) -> Autodesk.Revit.DB.FilteredElementCollector
    :return: A dictionary where key is the family name and values are ids of types belonging to that family.
    :rtype: Dictionary {str:[Autodesk.Revit.DB.ElementId]}
    """

    # get all Wall Type Elements
    wts = type_getter(doc)
    used_wts = {}
    used_wts = build_type_dictionary(wts, used_wts)
    return used_wts



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to deleting Revit shared parameters.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Utilities.Objects import result as res
from duHast.Utilities import utility as util
from duHast.Revit.SharedParameters.shared_parameters import (
    check_whether_shared_parameters_are_in_file,
    get_all_shared_parameters,
)
from duHast.Revit.Common import delete as rDel


def delete_shared_parameter_by_name(doc, shared_parameter_name):
    """
    Deletes a single shared parameter based on a name provided.

    param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param shared_parameter_name: The name of the shared parameter.
    :type shared_parameter_name: str
    :return: Result class instance.

        - Parameter delete status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the shared parameter.
        - result.status will be an empty list.
        
        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.status will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    paras = get_all_shared_parameters(doc)
    delete_ids = []
    parameter_name = "Unknown"
    for p in paras:
        if p.Name == shared_parameter_name:
            delete_ids.append(p.Id)
            # there should just be one match
            parameter_name = util.encode_ascii(rdb.Element.Name.GetValue(p))
            break
    if len(delete_ids) > 0:
        return_value = rDel.delete_by_element_ids(
            doc, delete_ids, "Delete Shared Parameter", parameter_name
        )
    else:
        return_value.update_sep(
            False,
            "parameter with guid: "
            + shared_parameter_name
            + " does not exist in file.",
        )
    return return_value


def delete_shared_parameter_by_guid(doc, guid):
    """
    Deletes a single shared parameter based on a guid provided.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param guid: A shared parameter GUID as string.
    :type guid: str
    :return: Result class instance.

        - Parameter delete status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the shared parameter.
        - result.status will be an empty list.

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.status will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    paras = get_all_shared_parameters(doc)
    delete_ids = []
    parameter_name = "Unknown"
    for p in paras:
        if p.GuidValue.ToString() == guid:
            delete_ids.append(p.Id)
            # there should just be one match
            parameter_name = util.encode_ascii(rdb.Element.Name.GetValue(p))
            break
    if len(delete_ids) > 0:
        return_value = rDel.delete_by_element_ids(
            doc, delete_ids, "Delete Shared Parameter", parameter_name
        )
    else:
        return_value.update_sep(
            False, "parameter with guid: {} does not exist in file.".format(guid)
        )
    return return_value


def delete_shared_parameters(doc, parameter_guids):
    """
    Deletes shared parameters by GUID from document.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameter_gui_ds: List of shared parameter GUIDs as string.
    :type parameter_gui_ds: list str
    :return: Result class instance.

        - Parameter deletion status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the shared parameter deleted.

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    one_got_deleted = False
    delete_guids = check_whether_shared_parameters_are_in_file(doc, parameter_guids)
    if len(delete_guids) > 0:
        for delete_guid in delete_guids:
            delete_status = delete_shared_parameter_by_guid(doc, delete_guid)
            # preserve TRUE value!
            if delete_status.status == True:
                one_got_deleted = True
            return_value.update(delete_status)
        return_value.update_sep(one_got_deleted, "Finished deleting parameters!")
    else:
        return_value.update_sep(True, "No matching shared parameters in file!")

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Functions around Revit BIM360.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
import clr

from duHast.Utilities.utility import get_local_app_data_path
from duHast.Utilities.files_get import get_files_from_directory_walker
from duHast.Utilities.files_io import get_file_size

# from System.IO import Path
from Autodesk.Revit.DB import ModelPathUtils


def get_bim_360_path(doc):
    """
    Gets human readable BIM 360 path.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: The path to the bim360 model. If an exception occurred an empty string will be returned.
    :rtype: str
    """

    # get bim 360 path
    revit_file_path = ""
    try:
        path = doc.GetCloudModelPath()
        revit_file_path = ModelPathUtils.ConvertModelPathToUserVisiblePath(path)
    except Exception as e:
        revit_file_path = ""
    return revit_file_path


def convert_bim_360_file_path(path):
    """
    Pretend this is a file server path rather than cloud model path and swap cloud path with C:/

    :param path: The model cloud file path starting with BIM360
    :type path: str

    :return: A file path without BIM360
    :rtype: str
    """

    # hack.. pretend path points to C:\\ rather than BIM 360:// or 'autodesk docs://'
    if path.lower().startswith("bim 360://"):
        # bim 360 naming
        path = r"C:/" + path[len("bim 360://") :]
    elif path.lower().startswith("autodesk docs://"):
        # autodesk construction cloud naming
        path = r"C:/" + path[len("autodesk docs://") :]
    return path


def get_model_bim_360_ids(doc):
    """
    Gets project id, model id, human readable path from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: project GUID, model GUID, human readable cloud path
    :rtype: GUID, GUID, str
    """

    path = doc.GetCloudModelPath()
    model_guid = path.GetModelGUID()
    project_guid = path.GetProjectGUID()
    # check whether this is a cloud model
    is_cloud_model = path.CloudPath
    # get human readable path
    human = ModelPathUtils.ConvertModelPathToUserVisiblePath(path)
    return project_guid, model_guid, str(human)


def get_model_file_size(doc):
    """
    Gets BIM360 file size, if file does not exists on local cache it will return -1.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: If file exists the file size in MB, otherwise -1
    :rtype: float
    """

    file_size = -1

    try:
        # path = doc.GetCloudModelPath()
        # full_path = rdb.ModelPathUtils.ConvertModelPathToUserVisiblePath(path)
        # get user environment
        host_name = get_local_app_data_path()
        # build path to local cache files
        folder = (
            host_name
            + "\\Autodesk\\Revit\\Autodesk Revit "
            + str(doc.Application.VersionNumber)
            + "\\CollaborationCache"
        )
        # local cache file name is same as file GUID on BIM360
        revit_file = doc.WorksharingCentralGUID.ToString()
        # get all files in cache folder matching GUID
        file_list = get_files_from_directory_walker(folder, revit_file)
        if len(file_list) > 0:
            for file in file_list:
                # just select one of the file instance..not to sure why this one?
                if file.Contains("CentralCache") == False:
                    file_size = get_file_size(file)
                    break
    except Exception as e:
        raise ValueError("Failed to get cloud model file size: {}".format(e))
    return file_size



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to Revit worksharing monitor process. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities import (
    files_csv as filesCSV,
    files_get as fileGet,
    files_io as fileIO,
)
from duHast.Utilities.Objects import result as res
from duHast.Utilities import system_process as sp


""" 
Process name of work sharing monitor
"""

PROCESS_NAME_WSM = "WorksharingMonitor.exe"
PROCESS_MARKER_FILENAME = "WSMProcessList"
PROCESS_MARKER_FILE_EXTENSION = ".plist"


# --------------------------------- worksharing monitor specific ---------------------------------------


def get_work_sharing_monitor_processes():
    """
    Get all currently running worksharing monitor processes

    :return: _description_
    :rtype: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    """
    all_processes = sp.get_all_running_processes()
    wsm_processes = sp.filter_by_process_name([PROCESS_NAME_WSM], all_processes)
    return wsm_processes


def write_out_wsm_data_to_file(directory_path):
    """
    Writes out all running worksharing monitor processes data to file

    :param directory_path: The directory path to where the marker file is to be saved.
    :type directory_path: str

    :return:
        Result class instance.

        - Export status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the fully qualified file path of the exported file.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    status = res.Result()
    process_list = get_work_sharing_monitor_processes()
    status_write_out = sp.write_out_process_data(
        directory_path,
        process_list,
        PROCESS_MARKER_FILENAME,
        PROCESS_MARKER_FILE_EXTENSION,
    )
    if status_write_out:
        status.update_sep(
            True,
            "Successfully wrote WSM process marker file to: " + str(directory_path),
        )
    else:
        status.update_sep(
            False, "Failed to write WSM process marker file to: " + str(directory_path)
        )
    return status


def delete_wsm_data_files(directory_path):
    """
    Deletes all WSM marker files in a directory.

    WSM marker files got a specific file extension: Check: PROCESS_MARKER_FILE_EXTENSION

    :param directory_path: The directory path containing marker files to be deleted.
    :type directory_path: str

    :return: True if all files where deleted successfully, otherwise False.
    :rtype: bool
    """

    status = True
    # get files in directory
    files_to_delete = fileGet.get_files_with_filter(
        directory_path, PROCESS_MARKER_FILE_EXTENSION
    )
    if len(files_to_delete) > 0:
        status_delete = True
        for file in files_to_delete:
            status_delete = status_delete and fileIO.file_delete(file)
    return status


def read_wsm_data_from_file(directory_path):
    """
    Reads all worksharing monitor processes data from marker file(s) in a given directory

    WSM marker files got a specific file extension: Check: PROCESS_MARKER_FILE_EXTENSION

    :param directory_path: The directory path to where marker files are to be read from.
    :type directory_path: str

    :return: list of list of str
    :rtype: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    """

    process_data = []
    files = fileGet.get_files_with_filter(directory_path, PROCESS_MARKER_FILE_EXTENSION)
    if len(files) > 0:
        for file in files:
            rows = filesCSV.read_csv_file(file)
            if len(rows) > 0:
                process_data = process_data + rows
    return process_data


def get_wsm_sessions_to_delete(ws_ms_to_keep):
    """
    Returns Worksharing monitor process sessions filtered by provided list (not in list)

    :param ws_ms_to_keep: List of worksharing monitor sessions to filter by. WSM included in this list will be removed from past in list.
    :type ws_ms_to_keep: List of list of str in format: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]

    :return: Filtered list of list of str of worksharing monitor sessions
    :rtype: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    """

    all_running_wsm_processes = get_work_sharing_monitor_processes()
    if len(ws_ms_to_keep) > 0:
        # get ids from list
        ids = []
        for wsm_p in ws_ms_to_keep:
            ids.append(wsm_p[3])
        filtered_processes = sp.filter_by_process_ids(ids, ws_ms_to_keep, False)
        return filtered_processes
    else:
        return all_running_wsm_processes


def clean_up_wsm_data_files(directory_path):
    """
    Removes all wsm data marker files in a given directory.

    WSM marker files got a specific file extension: Check: PROCESS_MARKER_FILE_EXTENSION.

    :param directory_path: The directory path containing the marker files to be deleted.
    :type directory_path: str

    :return:
        Result class instance.

        - Delete status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will confirm successful deletion of all files.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    status = res.Result()
    # attempt to delete old marker files
    status_delete = delete_wsm_data_files(directory_path)
    if status_delete:
        status.update_sep(True, "Successfully deleted WSM marker file(s)!")
    else:
        status.update_sep(False, "Failed to delete WSM marker file(s)!")
    return status


def die_wsm_die(directory_path, ignore_marker_files=False):
    """
    Kills all worksharing monitor processes currently active.

    Unless marker files are used. In that case only worksharing monitor sessions identified in marker files will be killed.

    :param directory_path: The directory path to where marker files are to be read from.
    :type directory_path: str
    :param ignore_marker_files: True no marker file data will be read and all WSM sessions running will be killed., defaults to False
    :type ignore_marker_files: bool, optional

    :return:
        Result class instance.

        - Kill status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will confirm successful killing of all WSM processes.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    status = res.Result()
    try:
        wsm_running_prior = []
        if ignore_marker_files == False:
            # read out worksharing monitor sessions running before script started
            wsm_running_prior = read_wsm_data_from_file(directory_path)
        wsm_to_delete = get_wsm_sessions_to_delete(wsm_running_prior)
        status_kill = sp.kill_processes(wsm_to_delete)
        if status_kill:
            status.update_sep(True, "All WSM sessions where killed.")
        else:
            status.update_sep(False, "Failed to kill all WSM sessions.")
    except Exception as e:
        status.update_sep(
            False, "Failed to kill wsm sessions with exceptions: " + str(e)
        )
    return status



# setup.py is required for local installation of the duHast package through:
# ipy -m pip install -e .

from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory/"README.md").read_text()

setup(
    package_dir={"": "./src"},
    packages=find_packages('./src'),
    long_description=long_description,
    long_description_content_type='text/markdown'
)